program jcPrepTraces
  ! Jason P. Chang
  ! March 2013
  !
  ! similar to jcBandpass1d, but with 4 axes
  ! fixed the whitening operation
  ! program algorithm:
  !   a) tapers the input trace in time
  !        topt: options for type of taper
  !               0) do nothing
  !               1) rectangle
  !               2) hann-band (default)
  !
  !   b) whitens the spectrum prior to bandpassing, if desired
  !
  !   c) bandpasses a trace in one-dimension
  !        fopt: options for type of filter
  !               0) do nothing
  !               1) rectangle
  !               2) hann-band
  !               3) hann-half (default)
  !        does FFT shifts before FFT and after IFFT
  !        sfftw: single precision (dfftw: double precision)
  !   d) downsample
  !        j1: standard j1 definition
  !
  ! outputs:
  !   a) result after taper "taper" (optional)
  !   b) filtered spectrum "filt" (optional)

  use sep
  implicit none

  include "fftw3.f"

  ! allocate variables
  integer                           :: ii,jj,kk
  real, parameter                   :: pi=3.1415926535

  ! trace parameters
  integer                           :: nt,n2,n3,n4
  integer                           :: j1,nout
  real                              :: ot,dt

  ! taper parameters
  integer                           :: topt,two_hour
  integer                           :: t1_ind,t2_ind,t3_ind,t4_ind
  real                              :: t1    ,t2    ,t3    ,t4
  real,allocatable,dimension (:)    :: taper

  ! Fourier parameters
  integer                           :: nf
  real                              :: of,df

  ! bandpass parameters
  integer                           :: fopt
  integer                           :: f1_ind,f2_ind,f3_ind,f4_ind
  real                              :: f1    ,f2    ,f3    ,f4
  real,allocatable,dimension (:)    :: filt
  integer*8                         :: fft_for,fft_inv

  real,allocatable,dimension (:)    :: datat,out,temp
  complex,allocatable, dimension(:) :: dataf

  ! whitening parameters
  integer                           :: flen,hflen
  real,allocatable,dimension(:)     :: ampl,avg
  logical                           :: white
  
  ! open I/O
  call sep_init()

  !----------------------------------
  ! trace information
  call from_history('n1',nt,3600*500*2)
  call from_history('d1',dt,.002)
  call from_history('o1',ot,0.)
  call from_history('n2',n2,1 )
  call from_history('n3',n3,1 )
  call from_history('n4',n4,1 )

  !----------------------------------
  ! taper parameters
  call from_param('topt',topt,2)
  call from_param('t1'    ,t1    ,ot                )
  call from_param('t1_ind',t1_ind,nint((t1-ot)/dt)+1)
  if (topt.eq.2) then
    call from_param ('t2'    ,t2    ,ot+1*nt/3*dt      )
    call from_param ('t2_ind',t2_ind,nint((t2-ot)/dt)+1)
  else
    call from_param('t2'    ,t2    ,ot+(nt-1)*dt      )
    call from_param('t2_ind',t2_ind,nint((t2-ot)/dt)+1)
  end if
  call from_param ('t3'    ,t3    ,ot+2*nt/3*dt)
  call from_param ('t3_ind',t3_ind,nint((t3-ot)/dt)+1)
  call from_param ('t4'    ,t4    ,ot+(nt-1)*dt)
  call from_param ('t4_ind',t4_ind,nint((t4-ot)/dt)+1)
  ! make sure taper indices make sense
  if (t1_ind.lt.1.) t1_ind= 1
  if (t2_ind.lt.1.) t2_ind= 1
  if (t3_ind.lt.1.) t3_ind= 1
  if (t4_ind.lt.1.) t4_ind= 1
  if (t1_ind.gt.nt) t1_ind=nt
  if (t2_ind.gt.nt) t2_ind=nt
  if (t3_ind.gt.nt) t3_ind=nt
  if (t4_ind.gt.nt) t4_ind=nt

  ! create the taper
  allocate(taper(nt)); taper=0.
  select case(topt)
    case(0)
       ! do nothing
    case(1)
       taper(t1_ind:t2_ind) = 1.
    case(2)
       do ii=t1_ind,t2_ind
         taper(ii) = 0.5*(1.-cos(pi*(ii-t1_ind)/real(t2_ind-t1_ind)))
       end do
       taper(t2_ind:t3_ind)=1.
       do ii=t3_ind,t4_ind
         taper(ii) = 0.5*(1.+cos(pi*(ii-t3_ind)/real(t4_ind-t3_ind)))
       end do
  end select

  if (exist_file('taper')) then
    call to_history('n1',nt,'taper')
    call to_history('d1',dt,'taper')
    call to_history('o1',ot,'taper')
    call to_history('n2',n2,'taper')
    call to_history('n3',n3,'taper')
    call to_history('n4',n4,'taper')
  end if

  !----------------------------------
  ! make sure that time samples is even numbered (bc going from real to complex)
  if (floor(real(nt)/2.).ne.ceiling(real(nt)/2.)) call erexit("n1 needs to be divisible by 2")

  ! calculate Fourier units
  nf = nt/2+1
  of = 0.
  df = 1./(nt*dt)

  ! filter parameters
  call from_param('fopt',fopt,3)
  call from_param('f1'    ,f1    ,of                )
  call from_param('f1_ind',f1_ind,nint((f1-of)/df)+1)
  if (fopt.eq.2) then
    call from_param('f2'    ,f2    ,of+1*nf/3*df      )
    call from_param('f2_ind',f2_ind,nint((f2-of)/df)+1)
  else
    call from_param('f2'    ,f2    ,of+(nf-1)*df      )
    call from_param('f2_ind',f2_ind,nint((f2-of)/df)+1)
  end if
  call from_param ('f3'    ,f3    ,of+2*nf/3*df      )
  call from_param ('f3_ind',f3_ind,nint((f3-of)/df)+1)
  call from_param ('f4'    ,f4    ,of+(nf-1)*df      )
  call from_param ('f4_ind',f4_ind,nint((f4-of)/df)+1)
  ! make sure filter indices make sense
  if (f1_ind.lt.1.) f1_ind= 1
  if (f2_ind.lt.1.) f2_ind= 1
  if (f3_ind.lt.1.) f3_ind= 1
  if (f4_ind.lt.1.) f4_ind= 1
  if (f1_ind.gt.nt) f1_ind=nt
  if (f2_ind.gt.nt) f2_ind=nt
  if (f3_ind.gt.nt) f3_ind=nt
  if (f4_ind.gt.nt) f4_ind=nt

  ! create the filter
  allocate(filt(nf)); filt=0.

  select case(fopt)
    case(0)
      ! do nothing
    case(1)
      filt(f1_ind:f2_ind) = 1.
    case(2)
      do ii=f1_ind,f2_ind
        filt(ii) = 0.5*(1.-cos(pi*(ii-f1_ind)/real(f2_ind-f1_ind)))
      end do
      filt(f2_ind:f3_ind)=1.
      do ii=f3_ind,f4_ind
        filt(ii) = 0.5*(1.+cos(pi*(ii-f3_ind)/real(f4_ind-f3_ind)))
      end do
    case(3)
      filt(1:f1_ind) = 1.
      do ii=f1_ind,f2_ind
        filt(ii) = 0.5*(1.+cos(pi*(ii-f1_ind)/real(f2_ind-f1_ind)))
      end do
  end select

  if (exist_file('filt')) then
    call to_history('n1', nf ,'filt')
    call to_history('d1', df ,'filt')
    call to_history('o1', of ,'filt')
    call to_history('n2', n2 ,'filt')
    call to_history('n3', n3 ,'filt')
    call to_history('n4', n4 ,'filt')
  end if

  !----------------------------------
  ! whitening paramters
  ! averaging filter total length
  call from_param('white',white,.false.)
  call from_param('flen',flen,21)
  hflen = flen/2

  !----------------------------------
  ! downsample parameter
  call from_param('j1',j1,1)
  nout = ceiling(real(nt)/j1)

  call to_history('n1',nout )
  call to_history('n2',n2   )
  call to_history('n3',n3   )
  call to_history('n4',n4   )
  call to_history('d1',j1*dt)
  call to_history('o1',ot   )

  call sep_close()
  !----------------------------------

  ! allocate
  allocate(datat(nt  ))
  allocate(dataf(nf  ))
  allocate(temp (nt/2))
  allocate(out  (nout))
  allocate(ampl (nf  ))
  allocate(avg  (nf  ))

  ! set up FFT plans
  call sfftw_plan_dft_r2c_1d(fft_for,nt,datat,dataf,FFTW_ESTIMATE)
  call sfftw_plan_dft_c2r_1d(fft_inv,nt,dataf,datat,FFTW_ESTIMATE)

  ! scale the taper?
  filt = filt/real(nt)

  ! NOW DO EVERYTHING
  do ii=1,n2*n3*n4
    ! read in the trace
    call sreed('in',datat,4*nt)


    !----------------------------------
    ! step one: taper
    datat = datat*taper

    if (exist_file('taper')) call srite('taper',datat,4*nt)

    !----------------------------------
    ! step two: bandpass
    ! FFT shift the data
    temp               = datat(1     :nt/2)
    datat(1     :nt/2) = datat(nt/2+1:nt  )
    datat(nt/2+1:nt  ) = temp

    ! FFT the data
    call sfftw_execute(fft_for)
    !write(0,*) "finished fft"

    ! whiten
    if (white) then
      avg = 0.
      ampl = 0.
      !do jj=1,nf
      !  if ( abs(dataf(jj).ne.0.) ) then
      !    dataf(jj) = dataf(jj)/abs(dataf(jj))
      !  else
      !    dataf(jj) = 0.
      !  end if
      !end do
      do jj=1,nf
        ampl(jj) = abs(dataf(jj))
      end do

     ! calculate running filter average for each frequency (need to consider BCs)
      ! set the initial
      jj = 1+hflen
      do kk=jj-hflen,jj+hflen
        avg(jj) = avg(jj) + (ampl(kk)/real(flen))
     end do

      ! take care of the rest of the middle part
      do jj=2+hflen,nf-hflen
        avg(jj) = avg(jj-1) - ((ampl(jj-hflen-1)-ampl(jj+hflen))/real(flen))
     end do

      ! now handle boundaries
      do jj=1,hflen
        do kk=1,jj+hflen
          avg(jj) = avg(jj) + (ampl(kk)/real(jj+hflen))
        end do
      end do
      do jj=nf-hflen+1,nf
        do kk=jj-hflen,nf
          avg(jj) = avg(jj) + (ampl(kk)/real(nf-jj+hflen+1))
        end do
      end do

      ! normalize by filtered value at given frequency
      do jj=1,nf
        dataf(jj) = dataf(jj)/avg(jj)
      end do
    end if

    ! filter
    dataf = dataf*filt
    if (exist_file('filt')) call srite('filt',abs(dataf*sqrt(real(nt))),4*nf)
    !write(0,*) "finished taper"

    ! IFFT the data
    call sfftw_execute(fft_inv)
    !write(0,*) "finished inverse fft"
    ! IFFT shift the data
    temp               = datat(nt/2+1:nt  )
    datat(nt/2+1:nt  ) = datat(1     :nt/2)
    datat(1     :nt/2) = temp

    ! downsample and write out the trace
    out = datat(::j1)
    !write(0,*)"downsampled to",size(out)
    call srite('out',out,4*nout)
  end do

  ! clean up?
  call sfftw_destroy_plan(fft_for,fft_inv)
  deallocate(datat,dataf,temp,out,ampl,avg)

end program
