program jcSI
  ! Program for cross-correlating noise
  ! Different from Xcorr in that station-receiver pairs are all put on the second axis
  ! Inputs:
  !   - fileR: "receivers" data
  !   - fileS: "source" data
  !   - nlag: number of samples to keep of negative and positive time lags
  ! Outputs
  !   - correlation result over lag time
  !
  ! Notes: should handle all sources and receivers, but cannot choose sources and receivers within this program
  !        clearly different, but at least something is being outputted
  !        look at the length of the input traces (ie. the calculation of nt)

  use sep

  implicit none
  include "fftw3.f"

  ! declare variables
  integer                             :: ii,jj,kk
  integer                             :: nr1,nr2,ns1,ns2,nlag,nw,nt
  real                                :: dr1,dr2,or1,or2,ds1,ds2,os1,os2,maxi
  real,allocatable,dimension (:)      :: fft_t
  real,allocatable,dimension (:,:)    :: dataR_t,dataOut_t,dataS_t
  complex,allocatable,dimension (:)   :: fft_w,dataS_w
  complex,allocatable,dimension (:,:) :: dataR_w,dataOut_w
  logical                             :: normalize
  integer*8                           :: fft_for,fft_inv

  ! open I/O
  call sep_init()

  ! normalization
  call from_param('normalize',normalize,.false.)

  ! get inputs from receiver side
  if (exist_file('R')) then
    call from_aux('R','n1',nr1,1 )
    call from_aux('R','n2',nr2,1 )
    call from_aux('R','d1',dr1,1.)
    call from_aux('R','d2',dr2,1.)
    call from_aux('R','o1',or1,1.)
    call from_aux('R','o2',or2,1.)
  else
    call seperr('Need file containing receivers')
  end if

  ! get inputs from source side
  if (exist_file('S')) then
    call from_aux('S','n1',ns1,1 )
    call from_aux('S','n2',ns2,1 )
    call from_aux('S','d1',ds1,1.)
    call from_aux('S','d2',ds2,1.)
    call from_aux('S','o1',os1,1.)
    call from_aux('S','o2',os2,1.)
  else
    call seperr('Need file containing virtual sources')
  end if

  ! check input lengths for FFT
  if (floor(real(nr1)/2.).ne.ceiling(real(nr1)/2.)) call seperr("FileR should be divisible by 2")
  if (floor(real(ns1)/2.).ne.ceiling(real(ns1)/2.)) call seperr("FileS should be divisible by 2")

  ! check that the inputs are the same length
  if (nr1.ne.ns1) call seperr('nr1 and ns1 are not the same length')
  if (dr1.ne.ds1) call seperr('dr1 and ds1 are not the same length')
  if (or1.ne.os1) call seperr('or1 and os1 do not start at the same time')

  ! calculate the output lag times
  call from_param('nlag',nlag,nr1)

  ! output parameters
  call to_history('n1',2*nlag    ) !is there a +1 here?
  call to_history('o1', -nlag*dr1)
  call to_history('d1',       dr1)
  call to_history('n2',nr2*ns2)
  call to_history('o2',or2)
  call to_history('d2',dr2)
  
  call sep_close()

  ! pad with zeros (apparently)
  nt = 2*ns1

  ! calculate Fourier units
  nw = nt/2+1

  ! allocate
  allocate(dataR_t(nt,nr2)); dataR_t=0.
  allocate(dataR_w(nw,nr2)); dataR_w=0.
  allocate(dataS_t(nt,ns2)); dataS_t=0.
  allocate(dataS_w(nw))    ; dataS_w=0.
  allocate(fft_t(nt))
  allocate(fft_w(nw))
  allocate(dataOut_t(nt,nr2))
  allocate(dataOut_w(nw,nr2))

  ! prepare the FFT plans
  call sfftw_plan_dft_r2c_1d(fft_for,nt,fft_t,fft_w,FFTW_ESTIMATE)
  call sfftw_plan_dft_c2r_1d(fft_inv,nt,fft_w,fft_t,FFTW_ESTIMATE)

  !--------------------------------------------------------------------!
  ! FFT the receivers
  do ii=1,nr2
    call sreed("R",dataR_t(1:nr1,ii),4*nr1)
  end do
  do ii=1,nr2
    fft_t = 0.
    fft_w = 0.
    fft_t = dataR_t(:,ii)
    call sfftw_execute(fft_for)
    dataR_w(:,ii) = fft_w
  end do
  
  !--------------------------------------------------------------------!
  ! FFT the sources and do the correlations
  do kk=1,ns2
    call sreed("S",dataS_t(1:ns1,kk),4*ns1)
  end do
  do kk=1,ns2
    fft_t = 0.
    fft_w = 0.
    fft_t = dataS_t(:,kk)
    call sfftw_execute(fft_for)
    dataS_w = fft_w

    ! multiply in the fourier domain
    do jj=1,nr2
      do ii=1,nw
        dataOut_w(ii,jj) = dataR_w(ii,jj) * conjg(dataS_w(ii))
      end do
    end do

    ! perform IFFT of output
    do jj=1,nr2
      fft_w = dataOut_w(:,jj)
      call sfftw_execute(fft_inv)
      dataOut_t(:,jj) = fft_t
    end do

    if (normalize) then
      do jj=1,nr2
        maxi = maxval(abs(dataOut_t(:,jj)))
        dataOut_t(:,jj) = dataOut_t(:,jj)/maxi
        call srite("out",dataOut_t(1-nlag+nt:nt,jj),4*nlag)
        call srite("out",dataOut_t(1:nlag,jj),4*nlag)
      end do
    else
      do jj=1,nr2
        call srite("out",dataOut_t(1-nlag+nt:nt,jj),4*nlag)
        call srite("out",dataOut_t(1:nlag,jj),4*nlag)
      end do
    end if

  end do

  ! clean up
  deallocate(dataR_t,dataR_w,dataS_t,dataS_w,fft_t,fft_w,dataOut_t,dataOut_w)


end program
