program jcSpectra1d
  ! Program to compute Fourier spectrum (uses FFTW)
  ! Sjoerd de Ridder, Stanford, 1/26/2012
  !
  ! opt:
  ! 1) amplitude spectrum (default)
  ! 2) power spectrum
  ! 3) real and imaginary parts (may be negative)
  !
  ! unit:
  ! 1) Hz (default)
  ! 2) Radians
  !
  ! true_t:
  ! T: use physical units padd to make symmetric*
  ! F: just stuff through FFT  (default)
  ! fftshift:
  ! T: fftshift data  (default)
  ! F: do not fftshift data
  ! 
  ! * use o1 and d1 from_param to redefine physical units
  !
  ! n2 needs to be divisible by 2 (for usr of r2c fftw)
  ! use true_t=1 to get around this (it padds one value).
  
  use sep
  implicit none
  
  ! Include for FFTW
  include "fftw3.f"

  ! Memory variables
  real, allocatable, dimension(:)     :: datat,temp
  complex,allocatable, dimension(:)   :: dataf
  
  ! Axis units
  integer                         :: n1,nt,nf,n2,n3,n4,n5,n6,b,e
  real                            :: d1,o1,dt,ot,df,of
  
  ! Spectrum type
  integer                         :: opt      ! 1: ampl    2: power
  integer                         :: funit    ! 1: Hz      2: Radians
  
  ! Misc
  integer                         :: ii,jj,kk,ll,mm
  real, parameter                 :: pi=3.1415926535
  logical                         :: verbose,true_t,fftshift
  
  integer*8                       :: fftplan
  
  ! Open I/O headerfiles 
  call sep_init()
  call from_param('verbose',verbose,.false.)
  call from_param('verb'   ,verbose,verbose)
  if (verbose) write(0,*) '-----------------------------------'
  if (verbose) write(0,*) '        COMPUTE 1D SPECTRA         '
  if (verbose) write(0,*) '-----------------------------------'
  if (verbose) write(0,*) '    i/o looping over traces        '
  
  ! Read from History
  call from_history("n1",n1,1 )
  call from_history("n2",n2,1 )
  call from_history("n3",n3,1 )
  call from_history("n4",n4,1 )
  call from_history("n5",n5,1 )
  call from_history("n6",n6,1 )
  call from_history("o1",o1,0.)
  call from_history("d1",d1,1.)
  call from_param  ("o1",o1,o1)
  call from_param  ("d1",d1,d1)
  
  ! Can I do r2c FFTW?
  if (floor(real(n1)/2.).ne.ceiling(real(n1)/2.)) call erexit("This would be a lot easier if n1 is divisible by 2")

  call from_param('opt'  ,opt  ,1)  
  call from_param('funit',funit,1)

  call from_param('true_t'  ,true_t  ,.false.)  
  call from_param('fftshift',fftshift,.true. )

  ! true_t:
  if (verbose) then
    if ( opt.eq.1 ) write(0,*) "amplitude spectrum"
    if ( opt.eq.2 ) write(0,*) "power spectrum"
    write(0,*) "(fftshift,true_t)", (/fftshift,true_t/) 
    if (     fftshift) write(0,*) "fftshift data"
    if (.not.fftshift) write(0,*) "do not fftshift data"
    if (     true_t  ) write(0,*) "use physical units padd to make symmetric"
    if (.not.true_t  ) write(0,*) "just stuff through FFT"
  end if

  ! Determine how to treat time axis
  if (true_t) then
    ot = minval( (/o1,-abs(o1+n1*d1)/) )
    dt = d1
    nt = -ot/d1*2
    b  = (o1-ot)/d1+1
    e  = b+n1-1
  else
    ot=  0.
    dt= d1
    nt= n1
    b =  1
    e  = b+n1-1
  end if

  if (verbose) write(0,*) 'Defining FT axis as (-t,t): ',ot,ot+(nt-1)*d1
  if (verbose) write(0,*) 'read data into indices: ',b,e,'out of',nt
  
  ! Fourier units
  of = 0.
  df = 1./(nt*dt)
  nf = nt/2+1
  if (funit.eq.2) df=2.*pi*df

  ! Write new History
  call to_history("n1"   ,nf)
  call to_history("d1"   ,df)
  call to_history("o1"   ,of)
  if ( opt .eq. 3 ) call to_history("esize", 8)

  ! Allocate memories for trace
  allocate(datat(nt),dataf(nf),temp(nt/2))
  ! Initialize fftw plan  
  call sfftw_plan_dft_r2c_1d(fftplan, nt,  datat, dataf, FFTW_ESTIMATE);


  if (verbose) write(0,*) '-------------------------------------------'
  !!!! Actual Kernel
  do jj=1,n2*n3*n4*n5*n6
    if (verbose) write(0,"(A,I7.1,A,I7.3)") 'trace: ',jj,' out of',n2*n3*n4*n5*n6
    
    datat = 0.; dataf=0.
    ! Read the data
    call sreed('in',datat(b:e),4*n1)
    
    ! fftshift the data
    if (fftshift) then
      temp               = datat(     1:nt/2)
      datat(     1:nt/2) = datat(nt/2+1:nt  )
      datat(nt/2+1:nt  ) = temp
    end if

    ! fft the data
    call sfftw_execute( fftplan )

    ! Scale the spectrum
    dataf = dataf / sqrt(real(nt))

    ! Write the data
    if (opt.eq.3) then
      call srite('out',   (dataf)    ,8*nf)
    else if (opt.eq.2) then
      call srite('out',abs((dataf))**2.,4*nf)
    else
      call srite('out',abs(dataf)    ,4*nf)
    end if
    
  end do

  ! Deallocate memory
  if (verbose) write(0,*) '-----------------------------------'
  call sfftw_destroy_plan(fftplan)
  deallocate(datat,dataf,temp)
  call exit(0)
  
end program
