program srSpectra1d
  ! Program to taper a trace
  ! Sjoerd de Ridder, Stanford, 1/25/2012
  !
  ! opt: (taper types 
  ! 0)  'zero'            --  needed: 'nothing' 
  ! 1)  'rect' (default)  --  needed: t1,t2
  ! 2)  'linear'          --  needed: t1,t2
  ! 3)  'pyramid'         --  needed: t1,t2
  ! 4)  'pyramid-band'    --  needed: t1,t2,t3,t4
  ! 5)  'hann'            --  needed: t1,t2
  ! 6)  'hann-half'       --  needed: t1,t2
  ! 7)  'hann-band'       --  needed: t1,t2,t3,t4
  ! 8)  'gaussian'        --  needed: t1,t2,sigma
  ! 9)  'gaussian_half'   --  needed: t1,t2,sigma
  ! 10) 'gaussian_block'  --  needed: t1,t2,t3,t4,sigma
  !
  ! t1,t2 default to spanning the axis.
  ! t1,t2,t3,t4 default to splitting the axis three way.
  !
  ! All tapers have the form of a band or lowcut,
  ! Notch tapers and high cut tapers are constructed
  ! with inv=1 option.
  ! Scale the data using parameter scale (default 1.). 
  ! Specify indices instead of physical units using t1_ind,t2_ind,...
  ! Inspect taper by supplying 'taper' as filename-parameter.
  
  use sep                      ! Main SEP module
  implicit none
  
  integer                         :: n1,n2,n3,n4,n5,n6
  real                            :: d1,o1
  real,allocatable, dimension(:)  :: data
  real,allocatable, dimension(:)  :: taper
  
  ! For Tapers
  integer                         :: t1_ind,t2_ind,t3_ind,t4_ind,hind
  real                            :: t1,    t2,    t3,    t4
  real                            :: scaler
  real                            :: sigma  ! For Gaussian taper
  integer                         :: opt
  logical                         :: inv

  ! Misc
  integer                         :: ii,jj
  real, parameter                 :: pi=3.1415926535
  logical                         :: verbose
  
  ! Open I/O headerfiles
  call sep_init()
  call from_param('verbose',verbose,.false.)
  call from_param('verb'   ,verbose,verbose)
  if (verbose) write(0,*) '-------------------------------------------'
  if (verbose) write(0,*) '        APPLY 1D TAPER TO DATA             '
  if (verbose) write(0,*) '-------------------------------------------'
  if (verbose) write(0,*) '        i/o looping over traces            '

  ! Read axis Details
  call from_history("n1", n1 , 1 )
  call from_history("n2", n2 , 1 )
  call from_history("n3", n3 , 1 )
  call from_history("n4", n4 , 1 )
  call from_history("n5", n5 , 1 )
  call from_history("n6", n6 , 1 )
  call from_history("d1", d1 , 1.)
  call from_history("o1", o1 , 0.)
  
  ! Read input parameters for taper, there is lots of options :-)
  call from_param ("opt"  ,  opt,   1    )
  if (opt.lt.0..or.opt.gt.10) opt =  1
  call from_param ("inv"  , inv, .false. )
  call from_param ("sigma",sigma,   1.   )
  call from_param ("scale",scaler,  1.   )

  ! Read Tie points from param
  call from_param ("t1"    ,t1    , o1          )
  call from_param ("t1_ind",t1_ind, nint((t1-o1)/d1)+1)
  
  if (opt.eq.4 .or. opt.eq.7 .or. opt.eq.10) then
  call from_param ("t2"    ,t2    , o1+1*n1/3*d1)
  call from_param ("t2_ind",t2_ind, nint((t2-o1)/d1)+1)
  else
  call from_param ("t2"    ,t2    , o1+(n1-1)*d1)
  call from_param ("t2_ind",t2_ind, nint((t2-o1)/d1)+1)
  end if
  
  call from_param ("t3"    ,t3    , o1+2*n1/3*d1)
  call from_param ("t3_ind",t3_ind, nint((t3-o1)/d1)+1)
  
  call from_param ("t4"    ,t4    , o1+(n1-1)*d1)
  call from_param ("t4_ind",t4_ind, nint((t4-o1)/d1)+1)

  ! Make sure they fall within range
  if (t1_ind.lt.1.) t1_ind= 1
  if (t2_ind.lt.1.) t2_ind= 1
  if (t3_ind.lt.1.) t3_ind= 1
  if (t4_ind.lt.1.) t4_ind= 1
  if (t1_ind.gt.n1) t1_ind=n1
  if (t2_ind.gt.n1) t2_ind=n1
  if (t3_ind.gt.n1) t3_ind=n1
  if (t4_ind.gt.n1) t4_ind=n1
  
  if (verbose) then
    if (opt.eq.4 .or. opt.eq.7 .or. opt.eq.10) then
      if (verbose) write(0,"(A,F,X,F,X,F,X,F)") ' taper tie values : ',t1    ,t2    ,t3    ,t4
      if (verbose) write(0,"(A,I,X,I,X,I,X,I)") ' taper tie indices: ',t1_ind,t2_ind,t3_ind,t4_ind
    else
      if (verbose) write(0,"(A,F,X,F)") ' taper tie values : ',t1    ,t2
      if (verbose) write(0,"(A,I,X,I)") ' taper tie indices: ',t1_ind,t2_ind
    end if
  end if
  
  
  !!! TAPER CREATION
  allocate(taper(n1)); taper=0.
  select case(opt)

    case(0)
      if (verbose) write(0,*) 'taper: 0.'

    case(1)
      if (verbose) write(0,*) 'taper: rectangle'
      taper(t1_ind:t2_ind)=1.

    case(2)
      if (verbose) write(0,*) 'taper: linear'
      taper(1:t1_ind)=1.
      do ii=t1_ind,t2_ind
        taper(ii)=1.-real(ii-t1_ind)/real(t2_ind-t1_ind)
      end do
      
    case(3)
      if (verbose) write(0,*) 'taper: pyramid'
      hind = nint(real(t2_ind-t1_ind)/2.)
      do ii=t1_ind,t1_ind+hind
        taper(ii)=   real(ii-t1_ind     )/real(t1_ind+hind-t1_ind)
      end do
      do ii=t1_ind+hind,t2_ind
        taper(ii)=1.-real(ii-t1_ind-hind)/real(t2_ind-t1_ind-hind)
      end do
      
    case(4)
      if (verbose) write(0,*) 'taper: pyramid-band'
      hind = nint(real(t2_ind-t1_ind)/2.)
      do ii=t1_ind,t2_ind
        taper(ii)=   real(ii-t1_ind     )/real(t2_ind-t1_ind)
      end do
      taper(t2_ind:t3_ind) = 1.
      do ii=t3_ind,t4_ind
        taper(ii)=1.-real(ii-t3_ind     )/real(t4_ind-t3_ind)
      end do
      
    case(5)
      if (verbose) write(0,*) 'taper: hann'
      do ii=t1_ind,t2_ind
        taper(ii)= .5*(1.-cos(2.*pi*(ii-t1_ind)/real(t2_ind-t1_ind)))
      end do
      
    case(6)
      if (verbose) write(0,*) 'taper: hann-half'
      taper(1:t1_ind) = 1.
      do ii=t1_ind,t2_ind
        taper(ii)= .5*(1.+cos(pi*(ii-t1_ind)/real(t2_ind-t1_ind)))
      end do
      
    case(7)
      if (verbose) write(0,*) 'taper: hann-band'
      do ii=t1_ind,t2_ind
        taper(ii)= .5*(1.-cos(pi*(ii-t1_ind)/real(t2_ind-t1_ind)))
      end do
      taper(t2_ind:t3_ind) = 1.
      do ii=t3_ind,t4_ind
        taper(ii)= .5*(1.+cos(pi*(ii-t3_ind)/real(t4_ind-t3_ind)))
      end do

    case(8)
      if (verbose) write(0,*) 'taper: gaussian'
      hind = nint(real(t2_ind-t1_ind))
      do ii=t1_ind,t2_ind
        taper(ii)= exp(-(ii-hind)**2./(2*d1*sigma**2.) )
      end do
      
    case(9)
      if (verbose) write(0,*) 'taper: gaussian-half'
      taper(1:t1_ind)=1.
      do ii=t1_ind,t2_ind
        taper(ii)= exp(-((ii-t1_ind)**2.)/(2*d1*sigma**2.) )
      end do
      
    case(10)
      if (verbose) write(0,*) 'taper: gaussian-band'
      do ii=t1_ind,t2_ind
        taper(ii)= exp(-((ii-t2_ind)**2.)/(2*d1*sigma**2.) )
      end do
      taper(t2_ind:t3_ind) = 1.
      do ii=t3_ind,t4_ind
        taper(ii)= exp(-((ii-t3_ind)**2.)/(2*d1*sigma**2.) )
      end do
  
  end select
  
  ! Invert taper to lowpass or notch
  if (inv) taper = 1.-taper
  
  ! Taper as output
  if (exist_file("taper")) then
    if (verbose) write(0,*) 'write taper to file'
    call to_history("n1", n1 ,"taper")
    call to_history("d1", d1 ,"taper")
    call to_history("o1", o1 ,"taper")
    call srite("taper",taper,4*n1)
  end if
  
  ! Close all I/O to headerfiles
  call sep_close ()
  

  allocate(data(n1))
  if (verbose) write(0,*) '-------------------------------------------'
  !!!! Actual Kernel
  do jj=1,n2*n3*n4*n5*n6
    if (verbose) write(0,"(A,I7.1,A,I7.3)") ' trace: ',jj,' out of',n2*n3*n4*n5*n6
    
    data = 0.
    ! Read the data
    call sreed('in' ,data,4*n1)
    
    ! Actual taper the data
    data = taper*data
    
    ! Write the data
    call srite('out',data,4*n1)
    
  end do
  
  
  ! Deallocate Memory
  if (verbose) write(0,*) '-------------------------------------------'
  deallocate(data,taper)
  call exit(0)

  
end program

